/*
 * $Id: crf.cpp,v 1.19 2008/12/05 16:11:32 tsuruoka Exp $
 */

#include "crf.h"
#include <cmath>
#include <cstdio>
#include <cfloat>
#include <set>
//#include "lbfgs.h"
#include <ext/hash_set>
//#include "OWLQN.h"
#include "common.h"


//#include "t2exp.h"

using namespace std;

const bool OUTPUT_MARGINAL_PROB = true;

//const double MIN_FEATURE_WEIGHT = 0.001; // prune features when saving

extern int push_stop_watch();

static CRF_Model * pointer_to_working_object = NULL; // this is not a good solution...

vector<bool> train_sample_used;


double
CRF_Model::FunctionGradient(const vector<double> & x, vector<double> & grad)
{
  assert(_fb.Size() == x.size());
  for (size_t i = 0; i < x.size(); i++) {
    _vl[i] = x[i];
    if (_vl[i] < -50) _vl[i] = -50;
    if (_vl[i] >  50) _vl[i] =  50;
  }

  double score = update_model_expectation();

  if (_sigma == 0) {
    for (size_t i = 0; i < x.size(); i++) {
      grad[i] = -(_vee[i] - _vme[i]);
    }
  } else {
    const double c = 1 / (_sigma * _sigma);
    for (size_t i = 0; i < x.size(); i++) {
      grad[i] = -(_vee[i] - _vme[i] - c * _vl[i]);
    }
  }

  return -score;
}

double
CRF_Model::FunctionGradientWrapper(const vector<double> & x, vector<double> & grad)
{
  return pointer_to_working_object->FunctionGradient(x, grad);
}


/*
struct CRFObjective : public DifferentiableFunction 
{
  const CRF_Model *crfmodel;
  const double l1weight;

  CRFObjective(const CRF_Model *p, double l1weight = 0) : crfmodel(p), l1weight(l1weight) { }

  double Eval(const DblVec& input, DblVec& gradient) {
    return crfmodel->FunctionGradientWrapper(input, gradient);
  }
};
*/

int
CRF_Model::perform_BFGS(const int max_iter)
{
  /*
  cerr << "invoking the OWLQN routine" << endl;
  pointer_to_working_object = this;
  int size = _vl.size();
  DblVec init(_vl), ans(size);
  double regweight = _inequality_width;
  double tol = 1e-4;
  int m = 10;
  DifferentiableFunction *obj = new CRFObjective(this, regweight);
  bool quiet = false;

  OWLQN opt(quiet);
  opt.Minimize(*obj, init, ans, regweight, tol, m);
  _vl = ans;

  return 0;


  pointer_to_working_object = this;

  const int dim = _fb.Size();
  vector<double> x0(dim);

  for (int i = 0; i < dim; i++) { x0[i] = _vl[i]; }

  vector<double> x;
  if (_inequality_width > 0) {
    cerr << "performing OWL-QN" << endl;
    x = perform_OWLQN(CRF_Model::FunctionGradientWrapper, x0, _inequality_width, max_iter);
  } else {
    cerr << "performing L-BFGS" << endl;
    x = perform_LBFGS(CRF_Model::FunctionGradientWrapper, x0);
  }

  for (int i = 0; i < dim; i++) { _vl[i] = x[i]; }

  return 0;
  */
}


double CRF_Model::forward_prob(const int len)
{
  for (int x = 0; x < len; x++) {
    double total = 0;
    for (int i = 0; i < _num_classes; i++) {
      double sum = 0;
      if (x == 0) {
	sum = 1;
      } else {
	for (int j = 0; j < _num_label_clusters; j++) {
	  //	  sum += edge_weight(cluster2label[j], i) * forward_cache_clustered(x-1, j);
	  sum += edge_weight(j, label2cluster[i]) * forward_cache_clustered(x-1, j);
	}
      }
      sum *= state_weight(x, i);
      forward_cache(x, i) = sum;
      //      maxv = max(sum, maxv);
      total += sum;
    }
    for (int i = 0; i < _num_label_clusters; i++) forward_cache_clustered(x, i) = 0;

    for (int i = 0; i < _num_classes; i++) {
      forward_cache(x, i) /= total;
      forward_cache_clustered(x, label2cluster[i]) += forward_cache(x, i);

      state_weight(x, i) /= total;
    }
  }
  double total = 0;
  for (int i = 0; i < _num_classes; i++) {
    total += forward_cache(len-1, i);
  }

  //  cerr << total << " ";
  return total;
}


double CRF_Model::backward_prob(const int len)
{
  for (int x = len - 1; x >= 0; x--) {
    double total = 0;
    for (int i = 0; i < _num_classes; i++) {
      double sum = 0;
      if (x == len - 1) {
	sum = 1;
      } else {
	for (int j = 0; j < _num_label_clusters; j++) {
	  //	  sum += edge_weight(i, cluster2label[j]) * backward_cache_clustered(x+1, j);
	  sum += edge_weight(label2cluster[i], j) * backward_cache_clustered(x+1, j);
	}
      }
      sum *= state_weight(x, i);
      backward_cache(x, i) = sum;
      total += sum;
    }

    for (int i = 0; i < _num_label_clusters; i++) backward_cache_clustered(x, i) = 0;
    for (int i = 0; i < _num_classes; i++) {
      backward_cache_clustered(x, label2cluster[i]) += backward_cache(x, i);
    }

  }
  double total = 0;
  for (int i = 0; i < _num_classes; i++) {
    total += backward_cache(0, i);
  }

  //  cerr << total << " ";
  return total;
}


void CRF_Model::initialize_edge_weights()
{
  /*
  for (int i = 0; i < _label_bag.Size(); i++) {
    for (int j = 0; j < _label_bag.Size(); j++) {
      const int id = edge_feature_id(i, j);
      //      if (id < 0) { edge_weight[i][j] = 1; continue; }
      assert(id >= 0);
      const double ew = _vl[id];
      edge_weight(i, j) = exp(ew);
    }
  }
  */
  for (int i = 0; i < _label_cluster_bag.Size(); i++) {
    for (int j = 0; j < _label_cluster_bag.Size(); j++) {
      //      int il = cluster2label[i];
      //      int jl = cluster2label[j];
      const int id = edge_feature_id(i, j);
      //      if (id < 0) { edge_weight[i][j] = 1; continue; }
      assert(id >= 0);
      const double ew = _vl[id];
      //      edge_weight(il, jl) = exp(ew);
      edge_weight(i, j) = exp(ew);
    }
  }

}


void CRF_Model::initialize_state_weights(const Sequence & seq)
{
  vector<double> powv(_num_classes);
  for (size_t i = 0; i < seq.vs.size(); i++) {
    powv.assign(_num_classes, 0);
    const Sample & s = seq.vs[i];
    for (vector<int>::const_iterator j = s.positive_features.begin(); j != s.positive_features.end(); j++){
      for (vector<int>::const_iterator k = _feature2mef[*j].begin(); k != _feature2mef[*j].end(); k++) {
	const double w = _vl[*k];
	powv[_fb.Feature(*k).label()] += w;
      }
    }

    for (int j = 0; j < _num_classes; j++) {
      state_weight(i, j) = exp(powv[j]);
    }
  }
}

double CRF_Model::forward_backward(const Sequence & seq)
{
  initialize_state_weights(seq);

  //  num_tags = _num_classes;

  const double fp = forward_prob(seq.vs.size());
  const double bp = backward_prob(seq.vs.size());
  assert(abs(fp - 1) < 0.01);
  assert(abs(bp - 1) < 0.01);

  if (!(fp > 0 && fp < DBL_MAX && bp > 0 && bp < DBL_MAX)) {
    cerr << endl << "error: line:" << _line_counter << " floating overflow. a different value of Gaussian prior might work." << endl;
    exit(1);
    ////cerr << endl << "error: floating overflow.  " << endl;
    return 1.f; // may cause assert failure because forward_cache/backward_cache are also overflow
  }

  //  assert(abs(fp - bp) < 0.1);
  assert(fp > 0 && fp < DBL_MAX);
  assert(bp > 0 && bp < DBL_MAX);

  return fp;
}



double CRF_Model::viterbi(const Sequence & seq, vector<int> & best_seq)
{
  initialize_state_weights(seq);

  //  num_tags = _num_classes;

  int cluster_max[CRF_Model::MAX_LABEL_TYPES];

  const int len = seq.vs.size();

  for (int x = 0; x < len; x++) {

    double total = 0;
    for (int i = 0; i < _num_classes; i++) {
      double m = -DBL_MAX;
      if (x == 0) {
	m = 1;
      } else {
	/*
	for (int j = 0; j < _num_classes; j++) {
	  double score = edge_weight(j, i) * forward_cache(x-1, j);
	  if (score > m) {
	    m = score;
	    backward_pointer(x, i) = j;
	  }
	}
	*/
	for (int j = 0; j < _num_label_clusters; j++) {
	  //	  const double score = edge_weight(cluster2label[j], i) * forward_cache(x-1, cluster_max[j]);
	  const double score = edge_weight(j, label2cluster[i]) * forward_cache(x-1, cluster_max[j]);
	  if (score > m) {
	    m = score;
	    backward_pointer(x, i) = cluster_max[j];
	  }
	}

      }
      m *= state_weight(x, i);
      forward_cache(x, i) = m;

      total += m;
    }
    for (int i = 0; i < _num_classes; i++) {
      forward_cache(x, i) /= total;
    }

    //    cluster_max.assign(max_cluster.size(), -1);
    for (int j = 0; j < _num_label_clusters; j++) cluster_max[j] = cluster2label[j];
    for (int i = 0; i < _num_classes; i++) {
      int & cm = cluster_max[label2cluster[i]];
      if (forward_cache(x, i) > forward_cache(x, cm)) cm = i;
    }
  }

  double m = -DBL_MAX;
  for (int i = 0; i < _num_classes; i++) {
    double score = forward_cache(len-1, i);
    if (score > m) {
      m = score;
      best_seq[len-1] = i;
    }
  }
  for (int x = len - 2; x >= 0; x--) {
    best_seq[x] = backward_pointer(x+1, best_seq[x+1]);
  }

  return 0;
}

void CRF_Model::nbest_search(const double lb, const int len, const int x, const int y,
                         const double rhs_score, vector<Path> & vp)
{
  if (x < len && forward_cache(x, y) * rhs_score < lb) return;

  nbest_search_path[x] = y;

  // root node
  if (x == len) {
    for (int i = 0; i < _num_classes; i++) {
      nbest_search(lb, len, x - 1, i, rhs_score, vp);
    }
    return;
  }

  const double sw = state_weight(x, y);

  // leaf nodes
  if (x == 0) {
    const double path_score = rhs_score * sw;
    Path p(path_score, vector<int>(&(nbest_search_path[0]), &(nbest_search_path[len])));
    vp.push_back(p);
    return;
  }

  for (int i = 0; i < _num_classes; i++) {
    //    const double ew = edge_weight(i, y);
    const double ew = edge_weight(label2cluster[i], label2cluster[y]);
    nbest_search(lb, len, x - 1, i, rhs_score * sw * ew, vp);
  }
}

double CRF_Model::nbest(const Sequence & seq, vector<Path> & vp,
			const int max_num, const double min_prob)
{
  //  num_tags = _num_classes;

  initialize_state_weights(seq);
  const double fp = forward_prob(seq.vs.size());
  assert(abs(fp - 1) < 0.01);

  const int len = seq.vs.size();

  for (int x = 0; x < len; x++) {
    for (int i = 0; i < _num_classes; i++) {
      double m = -DBL_MAX;
      if (x == 0) {
	m = 1;
      } else {
	for (int j = 0; j < _num_classes; j++) {
	  double score = edge_weight(label2cluster[j], label2cluster[i]) * forward_cache(x-1, j);
	  if (score > m) {
	    m = score;
	    backward_pointer(x, i) = j;
	  }
	}
      }
      m *= state_weight(x, i);
      forward_cache(x, i) = m;
    }
  }

  double m = -DBL_MAX;
  for (int i = 0; i < _num_classes; i++) {
    double score = forward_cache(len-1, i);
    if (score > m) {
      m = score;
    }
  }

  // n-best
  int iter = 0;
  //  for (double lb = 0.5 * m; lb >= min_prob; lb *= 0.5) {
  for (double lb = 0.3 * m;; lb *= 0.3) {
    vp.clear();
    nbest_search(max(lb, min_prob), len, len, 0, 1.0, vp);
    //    double sum = 0;
    //    for (vector<Path>::const_iterator i = vp.begin(); i != vp.end(); i++) 
    //      sum += i->score;
    //    if (abs(sum - 1.0) < 0.000000001) break;
    if (iter++ > 1000) break;
    if ((int)vp.size() >= max_num) break;
    if (lb <= min_prob) break;
  }
  //  cerr << iter << endl;
  sort(vp.begin(), vp.end());

  return vp.size();
}


double CRF_Model::calc_loglikelihood(const Sequence & seq)
{
  const int len = seq.vs.size();

  double logp = 0;
  for (int i = 0; i < len; i++) {
    logp += log(state_weight(i, seq.vs[i].label));
  }
  for (int i = 0; i < len-1; i++) {
    logp += log(edge_weight(label2cluster[seq.vs[i].label], label2cluster[seq.vs[i+1].label]));
  }
  
  return logp;
}


int
CRF_Model::make_feature_bag(const int cutoff)
{
#ifdef USE_HASH_MAP
  typedef __gnu_cxx::hash_map<mefeature_type, int> map_type;
#else    
  typedef std::map<mefeature_type, int> map_type;
#endif

  map_type count;
  if (cutoff > 0) {
    for (vector<Sequence>::const_iterator k = _vs.begin(); k != _vs.end(); k++) {
      for (vector<Sample>::const_iterator i = k->vs.begin(); i != k->vs.end(); i++) {
	for (vector<int>::const_iterator j = i->positive_features.begin(); j != i->positive_features.end(); j++) {
	  count[ME_Feature(i->label, *j).body()]++;
	}
      }
    }
  }

  for (vector<Sequence>::const_iterator k = _vs.begin(); k != _vs.end(); k++) {
    for (vector<Sample>::const_iterator i = k->vs.begin(); i != k->vs.end(); i++) {
      for (vector<int>::const_iterator j = i->positive_features.begin(); j != i->positive_features.end(); j++) {
	const ME_Feature feature(i->label, *j);
	if (cutoff > 0 && count[feature.body()] <= cutoff) continue;
	_fb.Put(feature);
      }
    }
  }

  init_feature2mef();
  
  return 0;
}

double
CRF_Model::heldout_likelihood()
{
  double logl = 0;
  int ncorrect = 0, total_len = 0;

  initialize_edge_weights();
  for (std::vector<Sequence>::const_iterator i = _heldout.begin(); i != _heldout.end(); i++) {
    total_len += i->vs.size();
    const double fp = forward_backward(*i);
    logl += calc_loglikelihood(*i);

    for (size_t j = 0; j < i->vs.size(); j++) {
      const Sample & s = i->vs[j];
      vector<double> wsum = calc_state_weight(j);
      if (s.label == max_element(wsum.begin(), wsum.end()) - wsum.begin()) ncorrect++;
    }

  }
  _heldout_error = 1 - (double)ncorrect / total_len;

  return logl /= _heldout.size();

}

vector<double> CRF_Model::calc_state_weight(const int i) const
{
  vector<double> wsum(_num_classes);
  for (int j = 0; j < _num_classes; j++) {
    wsum[j] = forward_cache(i, j) / state_weight(i, j) * backward_cache(i, j);
  }

  return wsum;
}

void
CRF_Model::add_sample_empirical_expectation(const Sequence & seq, vector<double> & vee)
{
  for (size_t i = 0; i < seq.vs.size(); i++) {
    for (vector<int>::const_iterator j = seq.vs[i].positive_features.begin(); j != seq.vs[i].positive_features.end(); j++){
      for (vector<int>::const_iterator k = _feature2mef[*j].begin(); k != _feature2mef[*j].end(); k++) {
	if (_fb.Feature(*k).label() == seq.vs[i].label) {
	  assert(*k >= 0 && *k < _vee.size());
	  _vee[*k] += 1.0;
	}
      }
    }
  }

  for (int i = 0; i < (int)seq.vs.size() - 1; i++) {
    //    const int c0 = seq.vs[i].label;
    //    const int c1 = seq.vs[i+1].label;
    const int c0 = label2cluster[seq.vs[i].label];
    const int c1 = label2cluster[seq.vs[i+1].label];
    _vee[edge_feature_id(c0, c1)] += 1.0;
  }
}

double
CRF_Model::add_sample_model_expectation(const Sequence & seq, vector<double>& vme, int & ncorrect)
{
  const double fp = forward_backward(seq);
  assert(abs(fp - 1.0) < 0.01);
  const double logl = calc_loglikelihood(seq);

  for (size_t i = 0; i < seq.vs.size(); i++) {
    const Sample & s = seq.vs[i];

    // model expectation (states)
    const vector<double> wsum = calc_state_weight(i);
    for (vector<int>::const_iterator j = s.positive_features.begin(); j != s.positive_features.end(); j++){
      for (vector<int>::const_iterator k = _feature2mef[*j].begin(); k != _feature2mef[*j].end(); k++) {
	vme[*k] += wsum[_fb.Feature(*k).label()];
      }
    }

    if (s.label == max_element(wsum.begin(), wsum.end()) - wsum.begin()) ncorrect++;

    if (i == seq.vs.size() - 1) continue;

    // model expectation (edges)
    for (int j = 0; j < _num_label_clusters; j++) {
      const double lhs = forward_cache_clustered(i, j);
      for (int k = 0; k < _num_label_clusters; k++) {
	const double rhs = backward_cache_clustered(i+1, k);
	assert(lhs != DBL_MAX && rhs != DBL_MAX);
	const double w = lhs * edge_weight(j, k) * rhs;
	vme[edge_feature_id(j, k)] += w;
      }
    }
  }

  return logl;
}

static double
l1norm(const vector<double>& v)
{
  double sum = 0;
  for (int i = 0; i < v.size(); i++) sum += abs(v[i]);
  return sum;
}


double
CRF_Model::update_model_expectation()
{
  double logl = 0;
  int ncorrect = 0, total_len = 0;

  _vme.resize(_fb.Size());
  for (int i = 0; i < _fb.Size(); i++) _vme[i] = 0;

  int nused = _vs.size();
  if (train_sample_used.size() > 0) {
    nused = 0;
    for (int i = 0; i < _vs.size(); i++) if (train_sample_used[i]) nused++;
  }

  initialize_edge_weights();
  int i = 0;
  for (vector<Sequence>::const_iterator n = _vs.begin(); n != _vs.end(); n++, i++) {
    if (train_sample_used.size() > 0 && train_sample_used[i] == false) continue;
    const Sequence & seq = *n;
    total_len += seq.vs.size();
    logl += add_sample_model_expectation(seq, _vme, ncorrect);
  }

  for (int i = 0; i < _fb.Size(); i++) {
    //    _vme[i] /= _vs.size();
    _vme[i] /= nused;
  }
  
  _train_error = 1 - (double)ncorrect / total_len;

  //  logl /= _vs.size();
  logl /= nused;


  double heldout_logl = 0;
  fprintf(stderr, "%4d logl = %8.3f acc = %6.4f ", _iter_counter++, logl, 1 - _train_error);
  if (_nheldout > 0) {
    heldout_logl = heldout_likelihood();
    fprintf(stderr, "  heldout_logl = %8.3f, acc = %6.4f\n", heldout_logl, 1 - _heldout_error);
  } else {
    cerr << endl;
  }

  if (_inequality_width > 0) {
    const double l1 = l1norm(_vl);
    const double f = logl - _inequality_width * l1;
    int nonzero = 0;
    for (size_t j = 0; j < _fb.Size(); j++)
      if (_vl[j] != 0) nonzero++;
    cerr << "f = " << f << "\t" << l1 << "\t" << nonzero << "\t" << heldout_logl << "\t" << 1 - _heldout_error << endl;
  }
    
  
  if (_sigma > 0) {
    const double c = 1/(2*_sigma*_sigma);
    for (int i = 0; i < _fb.Size(); i++) {
      logl -= _vl[i] * _vl[i] * c;
    }
    // bug?? what about L1?
  }

  //logl /= _vs.size();
  
  //  fprintf(stderr, "iter =%3d  logl = %10.7f  train_acc = %7.5f\n", iter, logl, (double)ncorrect/train.size());
  //  fprintf(stderr, "logl = %10.7f  train_acc = %7.5f\n", logl, (double)ncorrect/_train.size());

  return logl;
}

void
CRF_Model::add_training_sample(const CRF_Sequence & seq)
{
  if (seq.vs.size() >= MAX_LEN) {
    cerr << "error: sequence is too long.";
    exit(1);
  }
  if (seq.vs.size() == 0) {
    cerr << "warning: empty sentence" << endl;
    return;
  }
  assert(seq.vs.size() > 0);

  Sequence s1;
  for (vector<CRF_State>::const_iterator i = seq.vs.begin(); i != seq.vs.end(); i++) {
    Sample s;
    s.label = _label_bag.Put(i->label);
    if (s.label >= MAX_LABEL_TYPES - 2) {
      cerr << "error: too many types of labels." << endl;
      exit(1);
    }
    assert(s.label >= 0 && s.label < MAX_LABEL_TYPES);
    for (vector<string>::const_iterator j = i->features.begin(); j != i->features.end(); j++) {
      s.positive_features.push_back(_featurename_bag.Put(*j));
    }
    s1.vs.push_back(s);
  }
  _vs.push_back(s1);
}

void 
CRF_Model::map_label_to_cluster(const string & ls, const string & cs)
{
  const int l = _label_bag.Id(ls);
  if (l < 0) { cerr << "error: unknown label" << endl; exit(1); }
  const int c = _label_cluster_bag.Put(cs);
  assert(label2cluster[l] == -1 || label2cluster[l] == c);
  label2cluster[l] = c;
  cluster2label[c] = l;
}

int
CRF_Model::train(const OptimizationMethod method, const int cutoff,
		 const double sigma, const double widthfactor)
		 //		 const double Nsigma2, const double widthfactor)
{
  if (sigma > 0 && widthfactor > 0) {
  //  if (Nsigma2 > 0 && widthfactor > 0) {
    cerr << "error: Gausian prior and inequality modeling cannot be used together." << endl;
    return 0;
  }
  if (_vs.size() == 0) {
    cerr << "error: no training data." << endl;
    return 0;
  }
  if (_nheldout >= (int)_vs.size()) {
    cerr << "error: too much heldout data. no training data is available." << endl;
    return 0;
  }
  //if (_nheldout > 0) random_shuffle(_vs.begin(), _vs.end());

  if (_label_cluster_bag.Size() == 0) {
    cerr << "error: label cluster mapping not provided" << endl;
    exit(1);
  } else {
    for (int i = 0; i < _label_bag.Size(); i++) {
      //      cerr << _label_bag.Str(i) << "\t" << _label_cluster_bag.Str(label2cluster[i]) << endl;
    }
  }

  _num_classes = _label_bag.Size();
  _num_label_clusters = _label_cluster_bag.Size();

  for (int i = 0; i < _nheldout; i++) {
    _heldout.push_back(_vs.back());
    _vs.pop_back();
  }
  
  int total_len = 0;
  for (size_t i = 0; i < _vs.size(); i++) total_len += _vs[i].vs.size();
  
  //  _sigma = sqrt((double)total_len / Nsigma2);
  //  if (Nsigma2 == 0) _sigma = 0;
  _sigma = sigma;
  _inequality_width = widthfactor / _vs.size();
  
  if (cutoff > 0) cerr << "cutoff threshold = " << cutoff << endl;
  if (_sigma > 0) cerr << "Gaussian prior sigma = " << _sigma << endl;
    //    cerr << "N*sigma^2 = " << Nsigma2 << " sigma = " << _sigma << endl;
  if (widthfactor > 0)
    cerr << "widthfactor = " << widthfactor << endl;
  cerr << "preparing for estimation...";
  make_feature_bag(cutoff);
  //  _vs.clear();
  cerr << "done" << endl;
  cerr << "number of state types = " << _num_classes << endl;
  cerr << "number of state clusters = " << _num_label_clusters << endl;
  cerr << "number of samples = " << _vs.size() << endl;
  cerr << "number of heldout samples = " << _nheldout << endl;
  cerr << "number of features = " << _fb.Size() << endl;

  cerr << "calculating empirical expectation...";
  _vee.resize(_fb.Size());
  _vee.assign(_vee.size(), 0.0);

  train_sample_used.resize(_vs.size(), true);

  int count = 0;
  for (vector<Sequence>::const_iterator n = _vs.begin(); n != _vs.end(); n++, count++) {
    add_sample_empirical_expectation(*n, _vee);
  }

  for (int i = 0; i < _vee.size(); i++) _vee[i] /= _vs.size();

  cerr << "done" << endl;
  
  _vl.resize(_fb.Size());
  _vl.assign(_vl.size(), 0.0);


  _iter_counter = 0;
  switch (method) {
  case BFGS:
    perform_BFGS(); break;
  case PERCEPTRON:
    perform_AveragedPerceptron(); break;
  case SGD:
    perform_SGD(); break;
  }

  if (_inequality_width > 0) {
    int sum = 0;
    for (int i = 0; i < _fb.Size(); i++) {
      if (_vl[i] != 0) sum++;
    }
    cerr << "number of active features = " << sum << endl;
  }

  return 0;
}


void
CRF_Model::get_features(list< pair< pair<string, string>, double> > & fl)
{
  fl.clear();
  for (StrDic::const_Iterator i = _featurename_bag.begin();
       i != _featurename_bag.end(); i++) {
    for (int j = 0; j < _label_bag.Size(); j++) {
      const string label = _label_bag.Str(j);
      const string history = i.getStr();
      int id = _fb.Id(ME_Feature(j, i.getId()));
      if (id < 0) continue;
      fl.push_back( make_pair(make_pair(label, history), _vl[id]) );
    }
  }

}


bool
CRF_Model::load_from_file(const string & filename)
{
  FILE * fp = fopen(filename.c_str(), "r");
  if (!fp) {
    cerr << "error: cannot open " << filename << "!" << endl;
    return false;
  }
  cerr << "loading " << filename;

  _label_bag.Clear();
  _label_cluster_bag.Clear();
  _vl.clear();
  _featurename_bag.Clear();
  _fb.Clear();

  char buf[1024];
  while(fgets(buf, 1024, fp)) {
    string line(buf);
    if (line == "\n") break;
    line = line.substr(0, line.size()-1);
    string::size_type t1 = line.find_first_of('\t');
    string labelname = line.substr(0, t1);
    string clustername = line.substr(t1 + 1);

    const int label  = _label_bag.Put(labelname);
    const int cluster = _label_cluster_bag.Put(clustername);
    label2cluster[label] = cluster;
    cluster2label[cluster] = label;
  }

  while(fgets(buf, 1024, fp)) {
    string line(buf);
    string::size_type t1 = line.find_first_of('\t');
    string::size_type t2 = line.find_last_of('\t');
    string classname = line.substr(0, t1);
    string featurename = line.substr(t1 + 1, t2 - (t1 + 1) );
    float lambda;
    string w = line.substr(t2+1);
    sscanf(w.c_str(), "%f", &lambda);
      
    int label = _label_bag.Id(classname);
    if (featurename.size() >= 3 && featurename.substr(0, 3) == "->\t") label = _label_cluster_bag.Id(classname);

    assert(label >= 0);
    const int feature = _featurename_bag.Put(featurename);
    _fb.Put(ME_Feature(label, feature));
    _vl.push_back(lambda);
  }
  
  // for zero-wight edges
  for (int i = 0; i < _label_bag.Size(); i++) {
    for (int j = 0; j < _label_bag.Size(); j++) {
      const string & label1 = _label_bag.Str(j);
      const int l1 = _featurename_bag.Put("->\t" + label1);
      const int id = _fb.Id(ME_Feature(i, l1));
      if (id < 0) {
	_fb.Put(ME_Feature(i, l1));
	_vl.push_back(0);
      }
    }
  }

  //  _num_classes = _label_bag.Size() - 2;
  _num_classes = _label_bag.Size();
  _num_label_clusters = _label_cluster_bag.Size();

  init_feature2mef();
  initialize_edge_weights();

  fclose(fp);

  cerr << "...done" << endl;

  return true;
}

void
CRF_Model::init_feature2mef()
{
  _feature2mef.clear();
  for (int i = 0; i < _featurename_bag.Size(); i++) {
    vector<int> vi;
    for (int k = 0; k < _num_classes; k++) {
      int id = _fb.Id(ME_Feature(k, i));
      if (id >= 0) vi.push_back(id);
    }
    _feature2mef.push_back(vi);
  }

  //init edge feature_id[][]
  /*
  for (int i = 0; i < _label_bag.Size(); i++) {
    for (int j = 0; j < _label_bag.Size(); j++) {
      const int ic = label2cluster[i];
      const int jc = label2cluster[j];
      const string & label1 = _label_cluster_bag.Str(jc);
      const int l1 = _featurename_bag.Put("->\t" + label1);
      const int id = _fb.Put(ME_Feature(ic, l1));
      
      edge_feature_id(i, j) = id;
    }
  }
  */
  for (int i = 0; i < _label_cluster_bag.Size(); i++) {
    for (int j = 0; j < _label_cluster_bag.Size(); j++) {
      const string & label1 = _label_cluster_bag.Str(j);
      const int l1 = _featurename_bag.Put("->\t" + label1);
      const int id = _fb.Put(ME_Feature(i, l1));
      
      edge_feature_id(i, j) = id;
    }
  }

}

bool
CRF_Model::save_to_file(const string & filename, const double th) const
{
  FILE * fp = fopen(filename.c_str(), "w");
  if (!fp) {
    cerr << "error: cannot open " << filename << "!" << endl;
    return false;
  }
  cerr << "saving " << filename << "...";

  for (int i = 0; i < _label_bag.Size(); i++) {
    fprintf(fp, "%s\t%s\n", _label_bag.Str(i).c_str(), _label_cluster_bag.Str(label2cluster[i]).c_str());
  }
  fprintf(fp, "\n");

  //  for (MiniStringBag::map_type::const_iterator i = _featurename_bag.begin();
  for (StrDic::const_Iterator i = _featurename_bag.begin();
       i != _featurename_bag.end(); i++) {
    for (int j = 0; j < _label_bag.Size(); j++) {
      string label = _label_bag.Str(j);
      string history = i.getStr();
      int id = _fb.Id(ME_Feature(j, i.getId()));
      if (id < 0) continue;
      if (_vl[id] == 0) continue; // ignore zero-weight features
      if (abs(_vl[id]) < th) continue; // cut off low-weight features

      if (history.size() >= 3 && history.substr(0, 3) == "->\t") label = _label_cluster_bag.Str(j);

      fprintf(fp, "%s\t%s\t%f\n", label.c_str(), history.c_str(), _vl[id]);
    }
  }

  fclose(fp);

  cerr << "done" << endl;

  return true;
}

void CRF_Model::decode_forward_backward(CRF_Sequence & s0, 
					vector< map<string, double> > & tagp)
{
  if (s0.vs.size() >= MAX_LEN) {
    cerr << "error: sequence is too long." << endl;
    return;
  }

  Sequence seq;

  for (vector<CRF_State>::const_iterator i = s0.vs.begin(); i != s0.vs.end(); i++) {
    Sample s;
    for (vector<string>::const_iterator j = i->features.begin(); j != i->features.end(); j++) {
      const int id = _featurename_bag.Id(*j);
      if (id >= 0) s.positive_features.push_back(id);
    }
    seq.vs.push_back(s);
  }
  
  tagp.clear();
  forward_backward(seq);
  for (size_t i = 0; i < seq.vs.size(); i++) {
    vector<double> wsum = calc_state_weight(i);
    map<string, double> tp;
    if (OUTPUT_MARGINAL_PROB) {
      double sum = 0;
      for (vector<double>::const_iterator j = wsum.begin(); j != wsum.end(); j++) sum += *j;
      s0.vs[i].label = "";
      assert(abs(sum -1) < 0.01);
      double maxp = -1;
      string maxtag;
      for (size_t j = 0; j < wsum.size(); j++) {
	double p = wsum[j]/sum;
	if (p <= 0.001) continue;
	tp[_label_bag.Str(j).c_str()] = p;
	if (p > maxp) { maxp = p; maxtag = _label_bag.Str(j).c_str();}
      }
      tagp.push_back(tp);
      s0.vs[i].label = maxtag;
    } else {
      const int l = max_element(wsum.begin(), wsum.end()) - wsum.begin();
      s0.vs[i].label = _label_bag.Str(l);
    }
  }

}

void CRF_Model::decode_viterbi(CRF_Sequence & s0)
{
  //  num_tags = _num_classes;

  if (s0.vs.size() >= MAX_LEN) {
    cerr << "error: sequence is too long." << endl;
    return;
  }

  Sequence seq;

  for (vector<CRF_State>::const_iterator i = s0.vs.begin(); i != s0.vs.end(); i++) {
    Sample s;
    for (vector<string>::const_iterator j = i->features.begin(); j != i->features.end(); j++) {
      const int id = _featurename_bag.Id(*j);
      if (id >= 0) s.positive_features.push_back(id);
    }
    seq.vs.push_back(s);
  }
  
  vector<int> vs(seq.vs.size());
  viterbi(seq, vs);
  for (size_t i = 0; i < seq.vs.size(); i++) {
    s0.vs[i].label = _label_bag.Str(vs[i]);
  }
}


void 
CRF_Model::decode_nbest(CRF_Sequence & s0, vector<pair<double, vector<string> > > & sequences, const int num, const double min_prob)
{
  //  num_tags = _num_classes;

  if (s0.vs.size() >= MAX_LEN) {
    cerr << "error: sequence is too long." << endl;
    return;
  }

  Sequence seq;

  for (vector<CRF_State>::const_iterator i = s0.vs.begin(); i != s0.vs.end(); i++) {
    Sample s;
    for (vector<string>::const_iterator j = i->features.begin(); j != i->features.end(); j++) {
      const int id = _featurename_bag.Id(*j);
      if (id >= 0) s.positive_features.push_back(id);
    }
    seq.vs.push_back(s);
  }

  vector<int> vs(seq.vs.size());
  vector<Path> nb;
  nbest(seq, nb, num, min_prob);

  sequences.clear();
  if (nb.size() == 0) return;

  for (size_t i = 0; i < seq.vs.size(); i++) {
    s0.vs[i].label = _label_bag.Str(nb[0].vs[i]);
  }
  for (vector<Path>::const_iterator i = nb.begin(); i != nb.end(); i++) {
    if (i->score < min_prob) break;
    vector<string> vstr(i->vs.size());
    for (size_t j = 0; j < vstr.size(); j++) {
      vstr[j] = _label_bag.Str((i->vs)[j]);
    }
    sequences.push_back(pair<double, vector<string> >(i->score, vstr));
    if ((int)sequences.size() >= num) break;
  }
}


int
CRF_Model::perform_AveragedPerceptron()
{
  const int dim = _fb.Size();

  vector<double> wsum(dim, 0);

  initialize_edge_weights();

  //  const double a = 1.0;

  int iter = 0;
  while(iter < 5) {

    for (int i = 0; i < dim; i++) wsum[i] += _vl[i] * _vs.size();
    iter++;
      
    int error_num = 0;
    int rest = _vs.size();
    for (vector<Sequence>::const_iterator n = _vs.begin(); n != _vs.end(); n++, rest--) {

      const Sequence & seq = *n;
      vector<int> vs(seq.vs.size());
      viterbi(seq, vs);

      double Loss = 0;
      for (size_t i = 0; i < vs.size(); i++) {
	if (vs[i] != seq.vs[i].label) Loss += 1.0;
      }
      if (Loss == 0) continue;

      if (Loss > 0) error_num++;

      // single best MIRA (just to compute the coefficient)
      map<int, double> X;
      // state
      for (size_t i = 0; i < seq.vs.size(); i++) {
	const Sample & s = seq.vs[i];
	if (s.label == vs[i]) continue;
	for (vector<int>::const_iterator j = s.positive_features.begin(); j != s.positive_features.end(); j++){
	  for (vector<int>::const_iterator k = _feature2mef[*j].begin(); k != _feature2mef[*j].end(); k++) {
	    const int label = _fb.Feature(*k).label();
	    if (label == s.label) X[*k] += 1.0;
	    if (label == vs[i])   X[*k] -= 1.0;
	  }
	}
      }
      // edge
      for (int i = 0; i < seq.vs.size() - 1; i++) {
	//	const int eid0 = edge_feature_id(seq.vs[i].label, seq.vs[i+1].label);
	//	const int eid1 = edge_feature_id(vs[i], vs[i+1]);
	const int eid0 = edge_feature_id(label2cluster[seq.vs[i].label], label2cluster[seq.vs[i+1].label]);
	const int eid1 = edge_feature_id(label2cluster[vs[i]], label2cluster[vs[i+1]]);
	if (eid0 == eid1) continue;
	X[eid0] += 1.0;
	X[eid1] -= 1.0;
      }
      
      double wX = 0, X2 = 0;
      for (map<int, double>::const_iterator i = X.begin(); i != X.end(); i++) {
	wX += _vl[i->first] * i->second;
	X2 += i->second * i->second;
      }
      const double a = max(0.0, (Loss - wX) / X2);

      // state
      for (size_t i = 0; i < seq.vs.size(); i++) {
	const Sample & s = seq.vs[i];
	if (s.label == vs[i]) continue;

	for (vector<int>::const_iterator j = s.positive_features.begin(); j != s.positive_features.end(); j++){
	  for (vector<int>::const_iterator k = _feature2mef[*j].begin(); k != _feature2mef[*j].end(); k++) {
	    const int label = _fb.Feature(*k).label();
	    if (label == s.label) { _vl[*k] += a; wsum[*k] += a * rest; }
	    if (label == vs[i])   { _vl[*k] -= a; wsum[*k] -= a * rest; }
	  }
	}
      }

      // edge
      for (int i = 0; i < int(seq.vs.size()) - 1; i++) {
	const int ic  = label2cluster[seq.vs[i].label];
	const int ic1 = label2cluster[seq.vs[i+1].label];

	//	const int eid0 = edge_feature_id(seq.vs[i].label, seq.vs[i+1].label);
	const int eid0 = edge_feature_id(ic, ic1);
	double & w0 = _vl[eid0];
	w0 += a;
	wsum[eid0] += a * rest;
	//	edge_weight(seq.vs[i].label, seq.vs[i+1].label) = exp(w0);
	edge_weight(ic, ic1) = exp(w0);

	const int vc = label2cluster[vs[i]];
	const int vc1 = label2cluster[vs[i+1]];
	//	const int eid1 = edge_feature_id(vs[i], vs[i+1]);
	const int eid1 = edge_feature_id(vc, vc1);
	double & w1 = _vl[eid1];
	w1 -= a;
	wsum[eid1] -= a * rest;
	//	edge_weight(vs[i], vs[i+1]) = exp(w1);
	edge_weight(vc, vc1) = exp(w1);
      }

    }
    cerr << "iter = " << iter << " error_num = " << error_num << endl;
    if (error_num == 0) break;
  }

  for (int i = 0; i < dim; i++) { 
    _vl[i] = wsum[i] / (iter * _vs.size());
  }


  return 0;
}





/*
 * $Log: crf.cpp,v $
 * Revision 1.19  2008/12/05 16:11:32  tsuruoka
 * move temporary arrays to heap
 *
 * Revision 1.18  2008/12/04 17:45:00  tsuruoka
 * add USE_MANY_LABELS
 *
 * Revision 1.17  2008/12/04 12:27:38  tsuruoka
 * initialize_state_weights(): speedup
 *
 * Revision 1.16  2008/12/04 11:20:46  tsuruoka
 * speedup
 *
 * Revision 1.15  2008/10/19 23:40:32  tsuruoka
 * modify the default mode
 *
 * Revision 1.14  2008/10/19 22:50:45  tsuruoka
 * use subset empricial expectation
 * add pseudogradient, l1-ball projection
 *
 * Revision 1.13  2008/08/14 23:19:04  tsuruoka
 * add stochastic gradient descent
 *
 * Revision 1.12  2008/07/29 22:30:29  tsuruoka
 * limit the range of _v[i] to avoid overflow
 *
 * Revision 1.11  2008/04/12 17:55:50  tsuruoka
 * crf bugfix
 *
 * Revision 1.10  2007/08/13 17:30:43  tsuruoka
 * add averaged_perceptron and single best MIRA
 *
 * Revision 1.9  2007/08/07 06:44:29  hillbig
 * change overflow process
 *
 * Revision 1.8  2007/07/11 19:20:31  tsuruoka
 * replace calc_likelihood() with calc_loglikelihood()
 *
 * Revision 1.7  2007/07/05 13:21:04  tsuruoka
 * add StrDic
 *
 * Revision 1.6  2007/06/19 14:11:14  tsuruoka
 * change message
 *
 * Revision 1.5  2007/06/19 09:17:22  tsuruoka
 * check input length
 *
 * Revision 1.4  2007/06/18 09:09:57  tsuruoka
 * replace BLMVM with LBFGS and OWLQN
 *
 * Revision 1.3  2007/06/07 20:12:49  tsuruoka
 * add --nbest option
 *
 * Revision 1.2  2007/06/07 14:27:18  tsuruoka
 * add nbest output for CRF
 *
 * Revision 1.1.1.1  2007/05/15 08:30:35  kyoshida
 * stepp tagger, by Okanohara and Tsuruoka
 *
 */

