/*
 * $Id: learn.cpp,v 1.2 2007/06/08 08:25:43 tsuruoka Exp $
 */

#include <stdio.h>
#include <fstream>
#include <map>
#include <list>
#include <set>
#include <iostream>
#include <cfloat>
#include <sstream>
#include <iomanip>
#include <string>
#include "crf.h"
#include "common.h"

bool USE_SGD = true;
extern double SGD_ETA0;
extern double SGD_ITER;
extern double SGD_ALPHA;

using namespace std;

static string MODEL_FILE = "./model";

CRF_State crfstate(const vector<Token> &vt, int p);

void read_input(istream * ifile, vector<Sentence> & vs)
{
  string line;
  Sentence sentence;
  while (getline(*ifile,line)) {
    istringstream is(line.c_str());
    string str, pos, base, chunk;
    is >> str >> base >> pos >> chunk;
    if (str.size() == 0) {
      if (sentence.size() > 0) vs.push_back(sentence);
      sentence.clear();
      continue;
    }

    Token token(str, pos, "-");
    token.base = base;
    token.chunk = chunk;
    sentence.push_back(token);
    //    if (n++ % 10000 == 0) cerr << ".";
  }
  if (sentence.size() > 0) vs.push_back(sentence);

  //  cerr << endl << "num_sentences = " << vs.size() << endl;
}

void
crfapply(CRF_Model & m, vector<Sentence> & vs)
{
  for (vector<Sentence>::iterator i = vs.begin(); i != vs.end(); i++) {
    Sentence & s = *i;
    CRF_Sequence cs;
    for (size_t j = 0; j < s.size(); j++) {
      CRF_State mes = crfstate(s, j);
      cs.vs.push_back(mes);
    }
    m.decode_viterbi(cs);
    vector<int> vc;
    for (size_t j = 0; j < s.size(); j++) {
      //      cout << s[j].str << " " << s[j].pos << " " << s[j].gold << " " << cs.vs[j].label << endl;
      cout << s[j].str << "\t" << s[j].base << "\t" << s[j].pos << "\t" << s[j].chunk << "\t" << cs.vs[j].label << endl;
    }
    cout << endl;
  }
}


int main(int argc, char** argv)
{
  for (int i = 1; i < argc; i++) {
    string v = argv[i];
    if (v == "-m") {
      MODEL_FILE = argv[i+1];
      i++;
      continue;
    }
  }

  CRF_Model m;
  if (!m.load_from_file(MODEL_FILE)) {
    exit(1);
  }

  vector<Sentence> test;
  read_input(&cin, test);
  crfapply(m, test);


}



/*
 * $Log$
 */

