#include "crf.h"
#include "common.h"

using namespace std;

CRF_State
crfstate(const vector<Token> &vt, int p)
{
  CRF_State sample;

  const string w0 = vt[p].str;
  const string pos0 = vt[p].pos;
  const string c0 = vt[p].chunk;

  sample.label = vt[p].gold;

  string wm1 = "|", posm1 = "|", cm1 = "|";
  if (p > 0) {
    wm1   = vt[p-1].str;
    posm1 = vt[p-1].pos;
    cm1   = vt[p-1].chunk;
  }

  string wm2 = "|", posm2 = "|";
  if (p > 1) {
    wm2   = vt[p-2].str;
    posm2 = vt[p-2].pos;
  }

  string wp1 = "|", posp1 = "|", cp1 = "|";
  if (p < (int)vt.size() - 1) {
    wp1   = vt[p+1].str;
    posp1 = vt[p+1].pos;
    cp1   = vt[p+1].chunk;
  }

  string wp2 = "|", posp2 = "|";
  if (p < (int)vt.size() - 2) {
    wp2   = vt[p+2].str;
    posp2 = vt[p+2].pos;
  }

  sample.add_feature("W0_" + w0);
  sample.add_feature("W-1_" + wm1);
  sample.add_feature("W+1_" + wp1);
  sample.add_feature("W-2_" + wm2);
  sample.add_feature("W+2_" + wp2);

  sample.add_feature("W-10_" + wm1 + "_" + w0);
  sample.add_feature("W0+1_" + w0  + "_" + wp1);

  sample.add_feature("P-1_" + posm1);
  sample.add_feature("P0_"  + pos0);
  sample.add_feature("P+1_" + posp1);

  sample.add_feature("P-2-1_" + posm2 + "_" + posm1);
  sample.add_feature("P-10_"  + posm1 + "_" + pos0);
  sample.add_feature("P0+1_"  + pos0 + "_" + posp1);
  sample.add_feature("P+1+2_" + posp1 + "_" + posp2);

  sample.add_feature("P-2-10_" + posm2 + "_" + posm1 + "_" + pos0);
  sample.add_feature("P-10+1_" + posm1 + "_" + pos0  + "_" + posp1);
  sample.add_feature("P0+1+2_" + pos0  + "_" + posp1 + "_" + posp2);

  sample.add_feature("C-1_" + cm1);
  sample.add_feature("C0_"  + c0);
  sample.add_feature("C+1_" + cp1);

  char buf[1000];
  // substrings
  string lt = "|" + w0 + "|";
  for (int i = 0; i < lt.size(); i++) lt[i] = tolower(lt[i]);
  for (int len = 2; len < 10; len++) {
    for (int i = 0; i < (int)lt.size() - (len-1); i++) {
      const string ss = lt.substr(i, len);
      sprintf(buf, "SS_%s", ss.c_str());
      sample.add_feature(buf);
    }
  }

  string ws;
  int prec = 0;
  for (int i = 0; i < w0.size(); i++) {
    int c = w0[i];
    if (c >= 'A' && c <= 'Z') c = 'A';
    if (c >= 'a' && c <= 'z') c = 'a';
    if (isdigit(c)) c = '#';
    if (prec != c) ws += string(1, c);
    prec = c;
  }
  sample.add_feature("WORDSHAPE_" + ws);

  //  sample.add_feature("BASE_" + vt[p].base);

  /*
  // bag-of-words
  for (int j = p - 5; j <= p + 5; j++) {
    if (j == p) continue;
    if (j < 0 || j >= vt.size()) continue;
    sample.add_feature("BW5_" + vt[j].str);
  }
  */

  //  for (int j = 0; j < vt.size(); j++)
  //    cout << vt[j].str << " ";
  //  cout << endl;
  //  cout << i << endl;
  /*
  cout << sample.label << "\t";
  for (vector<string>::const_iterator j = sample.features.begin(); j != sample.features.end(); j++) {
      cout << *j << " ";
  }
  cout << endl;
  */
  
  return sample;
}

